
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

#ifndef PBRT_OCTREE_MOD_H
#define PBRT_OCTREE_MOD_H
// octree.h*
#include "pbrt.h"
#include "geometry.h"
// OctreeMod Declarations
template <class NodeData> struct OctNodeMod {
    OctNodeMod() {
	E = 0.0;
	area = 0.0;
	weight = 0.0;
	numChildren = 0;
	isLeaf = false;
	for (int i = 0; i < 8; ++i)
	    children[i] = NULL;
    }
    ~OctNodeMod() {
	for (int i = 0; i < 8; ++i)
	    delete children[i];
    }
    OctNodeMod *children[8];
    vector<NodeData> data;
    
    // data to be stored
    float weight;
    Point avgLoc;
    Spectrum E;
    float area;

    int numChildren;
    bool isLeaf;


};
template <class NodeData, class LookupProc> class OctreeMod {
public:
	// OctreeMod Public Methods
	OctreeMod(const BBox &b, int md = 16)
		: bound(b) {
		maxDepth = md;
		maxLeafArea = 0.0;
	}
	void Add(const NodeData &dataItem, const BBox &dataBound) {
		addPrivate(&root, bound, dataItem, dataBound,
			DistanceSquared(dataBound.pMin, dataBound.pMax));
	}
	void Lookup(const Point &p, const LookupProc &process) {
		if (!bound.Inside(p)) return;
		lookupPrivate(&root, bound, p, process);
	}

	void Config() {

	    configPrivate(&root,0);
	}
	OctNodeMod<NodeData> root;
	float maxLeafArea;
private:
	// OctreeMod Private Methods
	void addPrivate(OctNodeMod<NodeData> *node, const BBox &nodeBound,
		const NodeData &dataItem, const BBox &dataBound, float diag2,
		int depth = 0);
	void lookupPrivate(OctNodeMod<NodeData> *node, const BBox &nodeBound, const Point &P,
			const LookupProc &process);

	void configPrivate(OctNodeMod<NodeData> *node,int depth);
	// OctreeMod Private Data
	int maxDepth;
	BBox bound;
	//OctNodeMod<NodeData> root;
};
// OctreeMod Method Definitions

template <class NodeData, class LookupProc>
void  OctreeMod<NodeData, LookupProc>::configPrivate(OctNodeMod<NodeData> *node,int depth) {
    
    //Point avgLoc();
    //Spectrum E(0.0);
    //float area = 0.0;

    if (depth == maxDepth) {
	int numSamples = 0;
	//std::cout<<"in leaf!"<<std::endl;
	for (int i = 0; i < int(node->data.size()); i++){
	    //std::cout<<"here1"<<std::endl;
	    //  std::cout<<"leaf node->E : "<<node->data[i].E<<std::endl;
	    node->E += node->data[i].E * node->data[i].area;
	    //std::cout<<"here2"<<std::endl;
	    //node->avgLoc += node->data[i].avgLoc * node->data[i].E;
	    // float childweight = (node->data[i].E.c[0] + node->data[i].E.c[1] + node->data[i].E.c[2]) 
	    //	* node->data[i].area;
	    float childweight = node->data[i].E.y();
	    //std::cout<<"childweight: "<<childweight<<std::endl;
	    //std::cout<<"node->data[i].Area: "<<node->data[i].area<<std::endl;
	    //std::cout<<"node->data[i].E: "<<node->data[i].E<<std::endl;
	    node->weight += childweight;
	    node->avgLoc.x = node->avgLoc.x + node->data[i].avgLoc.x * childweight;
	    node->avgLoc.y = node->avgLoc.y + node->data[i].avgLoc.y * childweight; 
	    node->avgLoc.z = node->avgLoc.z + node->data[i].avgLoc.z * childweight;
	    //std::cout<<"child loc: "<<node->data[i].avgLoc.x<<" "<<node->data[i].avgLoc.y<<" "<<node->data[i].avgLoc.z<<std::endl;
	    //std::cout<<"child avg loc: "<<node->avgLoc.x<<" "<<node->avgLoc.y<<" "<<node->avgLoc.z<<std::endl;
	    node->area += node->data[i].area;
	    node->numChildren++;
	    numSamples++;
	    //std::cout<<"here3"<<std::endl;
	}
	maxLeafArea = max(maxLeafArea,node->area);	
	//std::cout<<"leaf area is: "<<node->area<<std::endl;
	if (node->area > 0.0)
	    node->E = node->E/ node->area;

	if (node->weight > 0.0) {
	    
	    node->avgLoc.x = node->avgLoc.x/node->weight;
	    node->avgLoc.y = node->avgLoc.y/node->weight;
	    node->avgLoc.z = node->avgLoc.z/node->weight;
	    
	    //  std::cout<<"here4"<<std::endl;
	}
	//std::cout<<"weighted child avg loc: "<<node->avgLoc.x<<" "<<node->avgLoc.y<<" "<<node->avgLoc.z<<std::endl;
	node->isLeaf = true;
    	return;
    }

    
    int numSamples = 0;
    for ( int i = 0; i < 8; i++ ) {
	OctNodeMod<NodeData> *childNode = node->children[i];
	//	std::cout<<"looping through children: "<<i<<" "<<depth<<std::endl;
	if (childNode != NULL) {
	    configPrivate(childNode,depth+1);
	    //std::cout<<"here5"<<std::endl;
	    //float weight = (node->data[i].E.r + node->data[i].E.g + node->data[i].E.b) 
	    //   * node->data[i].area;
	    node->E += childNode->E * childNode->area;
	    //  std::cout<<"in parent, childnode "<<i<<" e is: "<<childNode->E<<std::endl;
	    
	    float childNodeWeight = childNode->E.y();
	    node->weight += childNodeWeight;
	    //std::cout<<"in parent, childnode weight is: "<<childNodeWeight<<std::endl;
	    //std::cout<<"in parent, parent weightis: "<<node->weight<<std::endl;
	    //std::cout<<"here6"<<std::endl;
	    //std::cout<<"in parent,  child loc: "<<childNode->avgLoc.x<<" "<<childNode->avgLoc.y<<" "<<childNode->avgLoc.z<<std::endl;
	    node->avgLoc.x = node->avgLoc.x + childNode->avgLoc.x * childNodeWeight;
	    node->avgLoc.y = node->avgLoc.y + childNode->avgLoc.y * childNodeWeight; 
	    node->avgLoc.z = node->avgLoc.z + childNode->avgLoc.z * childNodeWeight;
	    //node->avgLoc.x += childNode.avgLoc.x*childNode.E;
	    node->area += childNode->area;
	    node->numChildren+= childNode->numChildren;
	    numSamples++;
	    //std::cout<<"in parent, num samples "<<numSamples<<std::endl;
	}
	//std::cout<<"here7"<<std::endl;
    }

    if (node->numChildren <= 8) {
	node->isLeaf = true;
    }
    if (node->area > 0.0) {
	node->E = node->E/ node->area;
    }
    // if (node->area > 0.0) {
    //	node->E  = node->E / float(numSamples);
	
    //    }
    //std::cout<<"parent E is : "<<node->E<<std::endl;
    if (node->weight > 0.0) {
	
	node->avgLoc.x = node->avgLoc.x/node->weight;
	node->avgLoc.y = node->avgLoc.y/node->weight;
	node->avgLoc.z = node->avgLoc.z/node->weight;
    }
    // std::cout<<"weighted parent avg loc: "<<node->avgLoc.x<<" "<<node->avgLoc.y<<" "<<node->avgLoc.z<<std::endl;
    //std::cout<<"here8"<<std::endl;
}

template <class NodeData, class LookupProc>
void OctreeMod<NodeData, LookupProc>::addPrivate(
		OctNodeMod<NodeData> *node, const BBox &nodeBound,
		const NodeData &dataItem, const BBox &dataBound,
		float diag2, int depth) {
	// Possibly add data item to current octree node
	if (depth == maxDepth ||
		DistanceSquared(nodeBound.pMin,
		                nodeBound.pMax) < diag2) {
		node->data.push_back(dataItem);
		return;
	}
	// Otherwise add data item to octree children
	Point pMid = .5 * nodeBound.pMin + .5 * nodeBound.pMax;
	// Determine which children the item overlaps
	bool over[8];
	over[0] = over[1] = over[2] =
	    over[3] = (dataBound.pMin.x <= pMid.x);
	over[4] = over[5] = over[6] =
	    over[7] = (dataBound.pMax.x  > pMid.x);
	over[0] &= (dataBound.pMin.y <= pMid.y);
	over[1] &= (dataBound.pMin.y <= pMid.y);
	over[4] &= (dataBound.pMin.y <= pMid.y);
	over[5] &= (dataBound.pMin.y <= pMid.y);
	over[2] &= (dataBound.pMax.y  > pMid.y);
	over[3] &= (dataBound.pMax.y  > pMid.y);
	over[6] &= (dataBound.pMax.y  > pMid.y);
	over[7] &= (dataBound.pMax.y  > pMid.y);
	over[0] &= (dataBound.pMin.z <= pMid.z);
	over[2] &= (dataBound.pMin.z <= pMid.z);
	over[4] &= (dataBound.pMin.z <= pMid.z);
	over[6] &= (dataBound.pMin.z <= pMid.z);
	over[1] &= (dataBound.pMax.z  > pMid.z);
	over[3] &= (dataBound.pMax.z  > pMid.z);
	over[5] &= (dataBound.pMax.z  > pMid.z);
	over[7] &= (dataBound.pMax.z  > pMid.z);
	for (int child = 0; child < 8; ++child) {
		if (!over[child]) continue;
		if (!node->children[child])
			node->children[child] = new OctNodeMod<NodeData>;
		// Compute _childBound_ for octree child _child_
		BBox childBound;
		childBound.pMin.x = (child & 4) ? pMid.x : nodeBound.pMin.x;
		childBound.pMax.x = (child & 4) ? nodeBound.pMax.x : pMid.x;
		childBound.pMin.y = (child & 2) ? pMid.y : nodeBound.pMin.y;
		childBound.pMax.y = (child & 2) ? nodeBound.pMax.y : pMid.y;
		childBound.pMin.z = (child & 1) ? pMid.z : nodeBound.pMin.z;
		childBound.pMax.z = (child & 1) ? nodeBound.pMax.z : pMid.z;
		
		//node->E += dataItem.E;
		
		addPrivate(node->children[child], childBound,
		           dataItem, dataBound, diag2, depth+1);
	}
}
template <class NodeData, class LookupProc>
void OctreeMod<NodeData, LookupProc>::lookupPrivate(
		OctNodeMod<NodeData> *node, const BBox &nodeBound,
		const Point &p, const LookupProc &process) {
    //for (u_int i = 0; i < node->data.size(); ++i)
    //		process(p, node->data[i]);

    float deltaW = node->area/(pow(p.x - node->avgLoc.x,2)
			       +pow(p.y - node->avgLoc.y,2)
			       +pow(p.z - node->avgLoc.z,2));

	// Determine which octree child node _p_ is inside
	Point pMid = .5f * nodeBound.pMin + .5f * nodeBound.pMax;
	int child = (p.x > pMid.x ? 4 : 0) +
		(p.y > pMid.y ? 2 : 0) + (p.z > pMid.z ? 1 : 0);
	if (node->children[child]) {
		// Compute _childBound_ for octree child _child_
		BBox childBound;
		childBound.pMin.x = (child & 4) ? pMid.x : nodeBound.pMin.x;
		childBound.pMax.x = (child & 4) ? nodeBound.pMax.x : pMid.x;
		childBound.pMin.y = (child & 2) ? pMid.y : nodeBound.pMin.y;
		childBound.pMax.y = (child & 2) ? nodeBound.pMax.y : pMid.y;
		childBound.pMin.z = (child & 1) ? pMid.z : nodeBound.pMin.z;
		childBound.pMax.z = (child & 1) ? nodeBound.pMax.z : pMid.z;
		lookupPrivate(node->children[child], childBound, p,
			process);
	}
}

#endif // PBRT_OCTREE_MOD_H
